#button-quicktext {
  list-style-image: url("resource://quicktext/skin/button.large.png");
  -moz-image-region: rect(0px 24px 24px 0px);
}

#button-quicktext:hover {
  -moz-image-region: rect(24px 24px 48px 0px);
}

#button-quicktext[disabled] {
  -moz-image-region: rect(48px 24px 72px 0px);
}

toolbar[iconsize="small"] #button-quicktext {
  list-style-image: url("resource://quicktext/skin/button.small.png");
  -moz-image-region: rect(0px 16px 16px 0px);
}

toolbar[iconsize="small"] #button-quicktext:hover {
  -moz-image-region: rect(16px 16px 32px 0px);
}

toolbar[iconsize="small"] #button-quicktext[disabled] {
  -moz-image-region: rect(32px 16px 48px 0px);
}

.quicktext-icon {
  list-style-image: url("resource://quicktext/skin/button.small.png");
  -moz-image-region: rect(16px 16px 32px 0px);
}

dialog[OS=WINNT] legend.insideTab {
  padding-right: 5px;
  padding-left: 5px;
  margin-bottom:5px;
  background-color: -moz-Field;
}

.textarea-container {
  margin:4px;
  overflow: hidden;  
}

textarea {
  margin: 0px;
  resize: none;
}

#quicktext-toolbar button {
  border:       0px;
  margin-left:  2px;
  margin-right: 2px;
  min-width:    auto;
}
